/*
 * Copyright (c) 1997 John Birrell <jb@cimlogic.com.au>.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by John Birrell.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY JOHN BIRRELL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * Modified and added to by Richard Seaman, Jr. <dick@tar.com>
 *
 * $Id: spinlock.c,v 1.4 1998/06/09 23:13:10 jb Exp $
 *
 */

#include <stdio.h>
#include <posix4/sched.h>
#include <unistd.h>
#include "pthread.h"
#if defined(LINUXTHREADS)
#include "internals.h"
#endif
#include <string.h>
#include <err.h>


/* Note:  There is a difference in behavior in FreeBSD betweeen
 * sched_yield (available only if _KPOSIX_PRIORITY_SCHEDULING is
 * configured in the kernel), and SYS_yield (generally available).
 * SYS_yield will set the thread priority to MAXPRI before calling
 * mi_switch.  sched_yield appears to leave the priority alone.

 * What really should happen with _spin is this:
 * 1) If we're SMP and the lock owner is running on another CPU,
 *    we should just "spin" without actually yielding at all.
 * 2) If we're not SMP, or if we're SMP and the lock owner is not
 *    executing on another CPU at the moment, we should yield but
 *    make sure that the lock owner has a priority at least equal
 *    to ours.  If it isn't, it should be raised or we will have
 *    priority inversion.  Unfortunatly, we don't have a syscall
 *    to do this directly.  Doing it with a series of syscalls
 *    would seem to be a real performance drag.
 */
#define _spin(owner) sched_yield()

/*
 * Lock a location for the running thread. Yield to allow other
 * threads to run if this thread is blocked because the lock is
 * not available. Note that this function does not sleep. It
 * assumes that the lock will be available very soon.
 */

#ifdef USE_RECURSIVE_SPINLOCK
/* XXXX This version will not work correctly until libc is patched */
void _spinlock(spinlock_t *lck)
{
	pthread_descr self = thread_self();

	/*
	 * Try to grab the lock and loop if another thread grabs
	 * it before we do.
	 */
	while(_atomic_lock(&lck->access_lock)) {

		/* Check if already locked by the running thread: */
		if (lck->lock_owner == (long) self) 
			warn ("_spinlock() called recursively and is deadlocking");

		/* spin, don't hog the bus when checking. */
		while (lck->access_lock)
			/*
			 * Note that under certain race conditions we can get
			 * here without lock_owner being set at all.  In this
			 * case lock_owner should be zero.
			 */
			_spin(lck->lock_owner);


	}

	/* The running thread now owns the lock: */
	lck->lock_owner = (long) self;
}

void _spinunlock(spinlock_t *lck)
{
	/* Make sure we zero the lock_owner before we release the lock */
        lck->lock_owner  = 0;
        lck->access_lock = 0;
}

#else
void _spinlock(spinlock_t *lck)
{
	/*
	 * Try to grab the lock and loop if another thread grabs
	 * it before we do.
	 */
	while(_atomic_lock(&lck->access_lock)) {

		/* spin, don't hog the bus when checking. */
		while (lck->access_lock)
			/*
			 * Note that under certain race conditions we can get
			 * here without lock_owner being set at all.  In this
			 * case lock_owner should be zero.
			 */
			_spin(0);


	}
}

void _spinunlock(spinlock_t *lck)
{
        lck->access_lock = 0;
}

#endif

/*
 * Lock a location for the running thread. Yield to allow other
 * threads to run if this thread is blocked because the lock is
 * not available. Note that this function does not sleep. It
 * assumes that the lock will be available very soon.
 *
 * This function checks if the running thread has already locked the
 * location, warns if this occurs and creates a thread dump before
 * returning.
 */
void _spinlock_debug(spinlock_t *lck, char *fname, int lineno)
{
        pthread_descr self = thread_self();
	/*
	 * Try to grab the lock and loop if another thread grabs
	 * it before we do.
	 */
	while(_atomic_lock(&lck->access_lock)) {
		/* Check if already locked by the running thread: */
		if (lck->lock_owner == (long) self) {
			char str[256];
			snprintf(str, sizeof(str), "Warning: Thread %i attempted to lock %p from %s (%d) which it had already locked in %s (%d)\n", (int)self, lck, fname, lineno, lck->fname, lck->lineno);
			_write(2,str,strlen(str));

		}
		/* spin, don't hog the bus when checking. */
		while (lck->access_lock && lck->lock_owner != (long) self)
			/*
			 * Note that under certain race conditions we can get
			 * here without lock_owner being set at all.  In this
			 * case lock_owner should be zero.
			 */
			_spin(lck->lock_owner);

	}

	/* The running thread now owns the lock: */
	lck->lock_owner = (long) self;
	lck->fname = fname;
	lck->lineno = lineno;
}

int _spin_lock(spinlock_t *lck)
{
	/*
	 * Try to grab the lock and loop if another thread grabs
	 * it before we do.
	 */
	while(_atomic_lock(&lck->access_lock)) {
		/* spin, don't hog the bus when checking. */
		while (lck->access_lock)
			/*
			 * Note that under certain race conditions we can get
			 * here without lock_owner being set at all.  In this
			 * case lock_owner should be zero.
			 */
			_spin(lck->lock_owner);

	}

	/* The running thread now owns the lock: */
	lck->lock_owner = (long)thread_self();
	return (1);
}

/* 
 * XXXX this isn't really quite right as a recursive spinlock, since
 * ww should have a lock count too.
 */
int _spin_lock_recursive (spinlock_t *lck)
{
	pthread_descr self = thread_self();
	/*
	 * Try to grab the lock and loop if another thread grabs
	 * it before we do.
	 */
	while(_atomic_lock(&lck->access_lock)) {

		/*
		 * Note that under certain race conditions we can get
		 * here without lock_owner being set at all.  In this
		 * case lock_owner should be zero.
		 */

		/* Check if already locked by the running thread: */
		if (lck->lock_owner == (long) self)
			return(1);

		/* spin, don't hog the bus when checking. */
		while (lck->access_lock && lck->lock_owner != (long) self)
			_spin(lck->lock_owner);


	}

	/* The running thread now owns the lock: */
	lck->lock_owner = (long) self;
	return (1);
}

int  _spin_unlock (spinlock_t *lck)
{
	/* Make sure we zero the lock_owner before we release the lock */
        lck->lock_owner  = 0;
        lck->access_lock = 0;
	return 0;
}

int  _spin_init (spinlock_t *lck)
{
        bzero (lck, sizeof (*lck));
	return 0;
}

int  _spin_destroy (spinlock_t *lck)
{
        return 0;
}

int  _spin_trylock (spinlock_t *lck)
{
        pthread_descr self = thread_self();

	if (_atomic_lock(&lck->access_lock)) {
		return 0;
	}

	/* The running thread now owns the lock: */
	lck->lock_owner = (long) self;
	return 1;
}

/* 
 * XXXX this isn't really quite right as a recursive spinlock, since
 * ww should have a lock count too.
 */
int  _spin_trylock_recursive (spinlock_t *lck)
{
        pthread_descr self = thread_self();

	if (_atomic_lock(&lck->access_lock)) {

		/* Check if already locked by the running thread: */
		if (lck->lock_owner == (long) self)
			return 1;
	        else
			/*
			 * Note that under certain race conditions we can get
			 * here without lock_owner being set at all.  In this
			 * case lock_owner should be zero.  It is possible, 
			 * though unlikely that we'll return a false negative
			 * here if the real lock_owner is self, and just hasn't
			 * been set.
			 */
		        return 0;

	}

	/* The running thread now owns the lock: */
	lck->lock_owner = (long) self;
	return 1;
}

#pragma weak pthread_spin_init=_spin_init
#pragma weak pthread_spin_lock=_spin_lock
#pragma weak pthread_spin_unlock=_spin_unlock
#pragma weak pthread_spin_destroy=_spin_destroy
#pragma weak pthread_spin_trylock=_spin_trylock
#pragma weak spin_init=_spin_init
#pragma weak spin_lock=_spin_lock
#pragma weak spin_unlock=_spin_unlock
#pragma weak spin_destroy=_spin_destroy
#pragma weak spin_trylock=_spin_trylock
