/*
 * Copyright (c) 1999 Richard Seaman, Jr.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Richard Seaman, Jr.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY RICHARD SEAMAN, Jr. AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "pthread.h"
#include "internals.h"

/* As long as we have fized stack spacing and alignment requirements
 * for stacks, its not desireable to set the stack_addr.  This is
 * not recommended, unless you're just setting the value to NULL.
 * Setting the stack_addr to NULL allows the thread manager to
 * choose its own stack_addr, which is strongly recommended.
 */
int pthread_attr_setstackaddr(pthread_attr_t *attr, void *stackaddr)
{
	if (attr == NULL)
		return (EINVAL);
	_pthread_stack_init();

	/* Make sure we're properly aligned */
	if (GET_TLS_FROM_STACK(stackaddr) != stackaddr)
		return (EINVAL);
	attr->stack_addr = stackaddr;
	return (0);
}

/* There is a limit to the stack size -- can't be more than _stackspacing
 * less the _tlspagesize, assuming a zero guard sizew.  If you're increasing
 * the stacksize above the default, you should first reduce the guardsize.
 * Note that if you've compiled with the VM_STACK option set, you will
 * get an autogrow stack, so there's not much harm in maxing out the
 * stacksize.  However, the minimum autogrow stack is SGROWSIZ (currently
 * 128K on an i386).  If you want a smaller stack than this, set it here
 * and you'll get a fixed size stack of the size requested.
 */
int pthread_attr_setstacksize(pthread_attr_t *attr, size_t stacksize)
{
	if (attr == NULL)
		return (EINVAL);
	_pthread_stack_init();

	/* Make sure stacksize isn't too big. It would be adviseable
	 * to call pthread_attr_setguardsize before calling this
	 * function if you're increasing the stack size..
	 */
	if (stacksize > _stackspacing - _tlspagesize - attr->guard_size)
		return (EINVAL);
	else if (stacksize < PTHREAD_MIN_STACK)
		return (EINVAL);
	attr->stack_size = stacksize; 
	return (0);
}

/* If you're increasing the guardsize, be sure you have reduced the
 * stack_size first if its the default size.
 */
int pthread_attr_setguardsize (pthread_attr_t *attr, size_t guardsize)
{
	if (attr == NULL)
		return (EINVAL);
	_pthread_stack_init();

	if (guardsize > _stackspacing - _tlspagesize - attr->stack_size)
		return (EINVAL);
	attr->guard_size = guardsize;
	return (0);
}

/* Unless you've set it, it will normally be NULL, which just allows
 * the thread manager to pick a stack_addr itself.
 */
int pthread_attr_getstackaddr(pthread_attr_t *attr, void **stackaddr)
{
	if (attr == NULL || stackaddr == NULL)
		return (EINVAL);
	*stackaddr = attr->stack_addr;
	return (0);
}

int pthread_attr_getstacksize(pthread_attr_t *attr, size_t *stacksize)
{
	if (attr == NULL || stacksize == NULL)
		return (EINVAL);
	*stacksize = attr->stack_size;
	return (0);
}

int pthread_attr_getguardsize (pthread_attr_t *attr, size_t *guardsize)
{
	if (attr == NULL || guardsize == NULL)
		return (EINVAL);
	*guardsize = attr->guard_size;
	return (0);
}

int _pthread_check_stackattr (pthread_attr_t *attr)
{
	if (attr == NULL)
		return (EINVAL);
	_pthread_stack_init();

	/* Make sure we're properly aligned */
	if (attr->stack_addr != NULL &&
	    GET_TLS_FROM_STACK(attr->stack_addr) != attr->stack_addr)
		return (EINVAL);

	if (attr->stack_size > _stackspacing - _tlspagesize - attr->guard_size)
		return (EINVAL);
	else if (attr->stack_size < PTHREAD_MIN_STACK)
		return (EINVAL);
	return (0);
}

int _pthread_set_stack_defaults (pthread_attr_t *attr)
{
	if (attr == NULL)
		return (EINVAL);
	_pthread_stack_init();

	attr->stack_addr = NULL;
	/* This sets stack_size to the max allowable. */
	attr->stack_size = _stackspacing - _defaultguardsize - _tlspagesize;
	attr->guard_size = _defaultguardsize;
	return (0);
}

void _pthread_stack_init(void)
{
	if (_stackspacing == 0)
		_pthread_setstackspacing(STACK_SIZE, GUARDSIZE);
}
