/*
 * Copyright (c) 1998 Richard Seaman, Jr.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Richard Seaman, Jr.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY RICHARD SEAMAN, Jr. AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

/*
Not all these includes may be necessary
*/

#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/aio.h>
#include <sys/mman.h>
#include <sys/fcntl.h>
#include <sys/uio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <stdarg.h>
#include <poll.h>
#include "pthread.h"
#include "internals.h"

#if 0
int     _aio_suspend(const struct aiocb * const acb_list[], int nacb_listent,
		     const struct timespec *tm);
int     _msync(void *addr, size_t len, int flags);
int     _poll(struct pollfd *fds, unsigned int nfds, int timeout);
int     _fsync(int);
int     _select(int, fd_set *, fd_set *, fd_set *, struct timeval *);
off_t   _lseek(int, off_t, int);
ssize_t _read(int, void *, size_t);
ssize_t _pread(int, void *, size_t, off_t);
int     _sigsuspend(const sigset_t *);
int     _open(const char *, int, ...);
int     _fcntl(int, int, ...);
ssize_t _readv(int, const struct iovec *, int);
ssize_t _writev(int, const struct iovec *, int);
pid_t   _wait4(pid_t, int *, int, struct rusage *);
ssize_t _write(int, const void *, size_t);
ssize_t _pwrite(int, const void *, size_t, off_t);
#endif

/* The following wrappers impement cancallation points */

int aio_suspend( const struct aiocb * const acb_list[], int nacb_listent,
               const struct timespec *tm)
{
        int ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _aio_suspend (acb_list, nacb_listent,tm);

	/* This is a cancellation point */
	pthread_setcanceltype  (oldtype, NULL);
	return (ret);
}

int
close(int fd)
{
	int ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _close(fd);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

int
fcntl(int fd, int cmd,...)
{
	int             flags = 0;
	int             oldfd;
	int             ret;
	va_list         ap;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	/* Initialise the variable argument list: */
	va_start(ap, cmd);

	/* Process according to file control command type: */
	switch (cmd) {
	case F_DUPFD:
		oldfd = va_arg(ap, int);
		ret = _fcntl(fd, cmd, oldfd);
		break;
	case F_SETFD:
	case F_SETFL:
		flags = va_arg(ap, int);
		ret = _fcntl(fd, cmd, flags);
		break;
	case F_GETFD:
	case F_GETFL:
		ret = _fcntl(fd, cmd, 0);
		break;
	default:
		/* Might want to make va_arg use a union */
		ret = _fcntl(fd, cmd, va_arg(ap, void *));
		break;
	}

	/* Free variable arguments: */
	va_end(ap);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

int
fsync(int fd)
{
	int             ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _fsync(fd);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

int
msync(void *addr, size_t len, int flags)
{
	int ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _msync (addr, len, flags);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

int
nanosleep(const struct timespec *rqtp, struct timespec * rmtp)
{
	int ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _nanosleep (rqtp, rmtp);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);

}

int
open(const char *path, int flags,...)
{
	int             fd;
	int             mode = 0;
	va_list         ap;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	/* Check if the file is being created: */
	if (flags & O_CREAT) {
		/* Get the creation mode: */
		va_start(ap, flags);
		mode = va_arg(ap, int);
		va_end(ap);
	}
	/* Open the file: */
	fd = _open(path, flags, mode);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (fd);
}

int 
poll(struct pollfd *fds, unsigned int nfds, int timeout)
{
	int ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _poll(fds, nfds, timeout);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

  	return (ret);
}

ssize_t
read(int fd, void *buf, size_t nbytes)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _read(fd, buf, nbytes);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

ssize_t
pread(int fd, void *buf, size_t nbytes, off_t offset)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _pread(fd, buf, nbytes, offset);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

int
sigsuspend(const sigset_t *sigmask)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _sigsuspend(sigmask);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

ssize_t
readv(int fd, const struct iovec * iov, int iovcnt)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _readv(fd, iov, iovcnt);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

pid_t
wait4(pid_t wpid, int *status, int options, struct rusage *rusage)
{
        pid_t ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _wait4 (wpid, status, options, rusage);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);

}

ssize_t
write(int fd, const void *buf, size_t nbytes)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _write(fd, buf, nbytes);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

ssize_t
pwrite(int fd, const void *buf, size_t nbytes, off_t offset)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _pwrite(fd, buf, nbytes, offset);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

ssize_t
writev(int fd, const struct iovec * iov, int iovcnt)
{
	int	ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _writev(fd, iov, iovcnt);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}


int 
select(int numfds, fd_set * readfds, fd_set * writefds,
       fd_set * exceptfds, struct timeval * timeout)
{
	int ret;
	int oldtype;

	/* This is a cancellation point */
	pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &oldtype);

	ret = _select(numfds, readfds, writefds, exceptfds, timeout);

	/* This is a cancellation point */
	pthread_setcanceltype (oldtype, NULL);

	return (ret);
}

